@@ -1,7 +1,7 @@
 /*
  * Soft:        multipath device mapper target autoconfig
  *
- * Version:     $Id: main.c,v 1.24 2005/06/08 19:39:37 agk Exp $
+ * Version:     $Id: main.c,v 1.26 2005/07/20 19:44:54 bmarzins Exp $
  *
  * Author:      Copyright (C) 2003 Christophe Varoqui
  *
@@ -173,19 +173,25 @@ print_path (struct path * pp, int style)
 	if (pp->dev_t)
 		printf("%-7s ", pp->dev_t);
 
-	switch (pp->state) {
-	case PATH_UP:
-		printf("[ready ]");
-		break;
-	case PATH_DOWN:
-		printf("[faulty]");
-		break;
-	case PATH_SHAKY:
-		printf("[shaky ]");
-		break;
-	default:
-		break;
+	if (conf->list > 1) {
+		switch (pp->state) {
+		case PATH_UP:
+			printf("[ready ]");
+			break;
+		case PATH_DOWN:
+			printf("[faulty]");
+			break;
+		case PATH_GHOST:
+			printf("[ghost ]");
+			break;
+		case PATH_SHAKY:
+			printf("[shaky ]");
+			break;
+		default:
+			break;
+		}
 	}
+
 	switch (pp->dmstate) {
 	case PSTATE_ACTIVE:
 		printf("[active]");
@@ -304,9 +310,6 @@ print_mp (struct multipath * mpp)
 		default:
 			break;
 		}
-		if (mpp->nextpg && mpp->nextpg == j + 1)
-			printf("[best]");
-
 		printf("\n");
 
 		vector_foreach_slot (pgp->paths, pp, i)
@@ -488,41 +491,38 @@ pathcount (struct multipath * mpp, int state)
 	return count;
 }
 
-/*
- * detect if a path is in the map we are about to create but not in the
- * current one (triggers a valid reload)
- * if a path is in the current map but not in the one we are about to create,
- * don't reload : it may come back latter so save the reload burden
- */
-static int
-pgcmp2 (struct multipath * mpp, struct multipath * cmpp)
+static void
+compute_pgid(struct pathgroup * pgp)
 {
-	int i, j, k, l;
+	struct path * pp;
+	int i;
+
+	vector_foreach_slot (pgp->paths, pp, i)
+		pgp->id ^= (long)pp;
+}
+
+static int
+pgcmp (struct multipath * mpp, struct multipath * cmpp)
+{
+	int i, j;
 	struct pathgroup * pgp;
 	struct pathgroup * cpgp;
-	struct path * pp;
-	struct path * cpp;
-	int found = 0;
+	int r = 0;
 
 	vector_foreach_slot (mpp->pg, pgp, i) {
-		vector_foreach_slot (pgp->paths, pp, j) {
-			vector_foreach_slot (cmpp->pg, cpgp, k) {
-				vector_foreach_slot (cpgp->paths, cpp, l) {
-					if (pp == cpp) {
-						found = 1;
-						break;
-					}
-				}
-				if (found)
-					break;
+		compute_pgid(pgp);
+
+		vector_foreach_slot (cmpp->pg, cpgp, j) {
+			if (pgp->id == cpgp->id) {
+				r = 0;
+				break;
 			}
-			if (found)
-				found = 0;
-			else
-				return 1;
+			r++;
 		}
+		if (r)
+			return r;
 	}
-	return 0;
+	return r;
 }
 
 static void
@@ -569,7 +569,7 @@ select_action (struct multipath * mpp, vector curmp)
 		mpp->action = ACT_RELOAD;
 		return;
 	}
-	if (pgcmp2(mpp, cmpp)) {
+	if (pgcmp(mpp, cmpp)) {
 		condlog(3, "different path group topology");
 		mpp->action = ACT_RELOAD;
 		return;
@@ -755,7 +755,7 @@ static void
 usage (char * progname)
 {
 	fprintf (stderr, VERSION_STRING);
-	fprintf (stderr, "Usage: %s\t[-v level] [-d] [-l|-ll]\n",
+	fprintf (stderr, "Usage: %s\t[-v level] [-d] [-l|-ll|-f|-F]\n",
 		progname);
 	fprintf (stderr,
 		"\t\t\t[-p failover|multibus|group_by_serial|group_by_prio]\n" \
@@ -769,6 +769,7 @@ usage (char * progname)
 		"\t-d\t\tdry run, do not create or update devmaps\n" \
 		"\t-l\t\tshow multipath topology (sysfs and DM info)\n" \
 		"\t-ll\t\tshow multipath topology (maximum info)\n" \
+		"\t-F\t\tflush a multipath device map\n" \
 		"\t-F\t\tflush all multipath device maps\n" \
 		"\t-p policy\tforce all maps to specified policy :\n" \
 		"\t   failover\t\t1 path per priority group\n" \
@@ -854,7 +855,7 @@ main (int argc, char *argv[])
 	vector curmp = NULL;
 	vector pathvec = NULL;
 	int i;
-	int di_flag;
+	int di_flag = 0;
 	int arg;
 	extern char *optarg;
 	extern int optind;
@@ -865,10 +866,9 @@ main (int argc, char *argv[])
 		exit(1);
 	}
 
-	if (dm_prereq(DEFAULT_TARGET, 1, 0, 3)) {
-		condlog(0, "device mapper prerequisites not met");
+	if (dm_prereq(DEFAULT_TARGET, 1, 0, 3))
 		exit(1);
-	}
+
 	if (sysfs_get_mnt_path(sysfs_path, FILE_NAME_SIZE)) {
 		condlog(0, "multipath tools need sysfs mounted");
 		exit(1);
@@ -876,7 +876,7 @@ main (int argc, char *argv[])
 	if (load_config(DEFAULT_CONFIGFILE))
 		exit(1);
 
-	while ((arg = getopt(argc, argv, ":qdl::Fi:M:v:p:")) != EOF ) {
+	while ((arg = getopt(argc, argv, ":qdl::Ffi:M:v:p:")) != EOF ) {
 		switch(arg) {
 		case 1: printf("optarg : %s\n",optarg);
 			break;
@@ -890,9 +890,11 @@ main (int argc, char *argv[])
 		case 'd':
 			conf->dry_run = 1;
 			break;
+		case 'f':
+			conf->remove = 1;
+			break;
 		case 'F':
-			dm_flush_maps(DEFAULT_TARGET);
-			goto out;
+			conf->remove = 2;
 			break;
 		case 'l':
 			conf->list = 1;
@@ -941,6 +943,19 @@ main (int argc, char *argv[])
 
 	}
 
+	if (conf->remove == FLUSH_ONE) {
+		if (conf->dev_type == DEV_DEVMAP)
+			dm_flush_map(conf->dev, DEFAULT_TARGET);
+		else
+			condlog(0, "must provide a map name to remove");
+
+		goto out;
+	}
+	else if (conf->remove == FLUSH_ALL) {
+		dm_flush_maps(DEFAULT_TARGET);
+		goto out;
+	}
+
 	/*
 	 * allocate core vectors to store paths and multipaths
 	 */
@@ -966,12 +981,15 @@ main (int argc, char *argv[])
 	/*
 	 * get a path list
 	 */
+	if (conf->dev)
+		di_flag = DI_WWID;
+
 	if (conf->list > 1)
 		/* extended path info '-ll' */
-		di_flag = DI_SYSFS | DI_CHECKER;
+		di_flag |= DI_SYSFS | DI_CHECKER;
 	else if (conf->list)
 		/* minimum path info '-l' */
-		di_flag = DI_SYSFS;
+		di_flag |= DI_SYSFS;
 	else
 		/* maximum info */
 		di_flag = DI_ALL;
